@@ -127,7 +127,8 @@ class S3File(io.IOBase):
     def seek(self, offset, whence=os.SEEK_SET):
         if whence not in (os.SEEK_CUR, os.SEEK_END, os.SEEK_SET):
             raise ValueError("invalid value for 'whence'")
-        return self._f.seek(offset, whence)
+        self._f.seek(offset, whence)
+        return self._f.tell()
 
     def seekable(self):
         return True
@@ -155,12 +156,14 @@ class S3File(io.IOBase):
     def write(self, b):
         if not self.__mode.writing:
             raise IOError('not open for reading')
-        return self._f.write(b)
+        self._f.write(b)
+        return len(b)
 
     def truncate(self, size=None):
         if size is None:
             size = self._f.tell()
-        return self._f.truncate(size)
+        self._f.truncate(size)
+        return size
 
 
 @contextlib.contextmanager
@@ -199,16 +202,22 @@ class S3FS(FS):
     `PyFilesystem <https://pyfilesystem.org>`_
 
     :param str bucket_name: The S3 bucket name.
-    :param str dir_path: The root directory within the S3 Bucker.
-        Defaults to "/""
-    :param str aws_access_key_id: The access key, or None to read the
-        key from standard configuration files.
-    :param str aws_secret_access_key: The secret key, or None to read
+    :param str dir_path: The root directory within the S3 Bucket.
+        Defaults to ``"/"``
+    :param str aws_access_key_id: The access key, or ``None`` to read
         the key from standard configuration files.
-    :param str aws_sssion_token:
-    :param str region: Option S3 region.
+    :param str aws_secret_access_key: The secret key, or ``None`` to
+        read the key from standard configuration files.
+    :param str endpoint_url: Alternative endpoint url (``None`` to use
+        default).
+    :param str aws_session_token:
+    :param str region: Optional S3 region.
     :param str delimiter: The delimiter to separate folders, defaults to
         a forward slash.
+    :param bool strict: When ``True`` (default) S3FS will follow the
+        PyFilesystem specification exactly. Set to ``False`` to disable
+        validation of destination paths which may speed up uploads /
+        downloads.
 
     """
 
@@ -256,8 +265,10 @@ class S3FS(FS):
                  aws_access_key_id=None,
                  aws_secret_access_key=None,
                  aws_session_token=None,
+                 endpoint_url=None,
                  region=None,
-                 delimiter='/'):
+                 delimiter='/',
+                 strict=True):
         _creds = (aws_access_key_id, aws_secret_access_key)
         if any(_creds) and not all(_creds):
             raise ValueError(
@@ -270,8 +281,10 @@ class S3FS(FS):
         self.aws_access_key_id = aws_access_key_id
         self.aws_secret_access_key = aws_secret_access_key
         self.aws_session_token = aws_session_token
+        self.endpoint_url = endpoint_url
         self.region = region
         self.delimiter = delimiter
+        self.strict = strict
         self._tlocal = threading.local()
         super(S3FS, self).__init__()
 
@@ -331,7 +344,13 @@ class S3FS(FS):
     @property
     def s3(self):
         if not hasattr(self._tlocal, 's3'):
-            self._tlocal.s3 = boto3.resource('s3')
+            self._tlocal.s3 = boto3.resource(
+                's3',
+                aws_access_key_id=self.aws_access_key_id,
+                aws_secret_access_key=self.aws_secret_access_key,
+                aws_session_token=self.aws_session_token,
+                endpoint_url=self.endpoint_url
+            )
         return self._tlocal.s3
 
     @property
@@ -341,7 +360,8 @@ class S3FS(FS):
                 's3',
                 aws_access_key_id=self.aws_access_key_id,
                 aws_secret_access_key=self.aws_secret_access_key,
-                aws_session_token=self.aws_session_token
+                aws_session_token=self.aws_session_token,
+                endpoint_url=self.endpoint_url
             )
         return self._tlocal.client
 
@@ -515,9 +535,10 @@ class S3FS(FS):
 
             return s3file
 
-        info = self.getinfo(path)
-        if info.is_dir:
-            raise errors.FileExpected(path)
+        if self.strict:
+            info = self.getinfo(path)
+            if info.is_dir:
+                raise errors.FileExpected(path)
 
         def on_close(s3file):
             """Called when the S3 file closes, to upload the data."""
@@ -543,9 +564,10 @@ class S3FS(FS):
         self.check()
         _path = self.validatepath(path)
         _key = self._path_to_key(_path)
-        info = self.getinfo(path)
-        if info.is_dir:
-            raise errors.FileExpected(path)
+        if self.strict:
+            info = self.getinfo(path)
+            if info.is_dir:
+                raise errors.FileExpected(path)
         self.client.delete_object(
             Bucket=self._bucket_name,
             Key=_key
@@ -587,9 +609,10 @@ class S3FS(FS):
 
     def getbytes(self, path):
         self.check()
-        info = self.getinfo(path)
-        if not info.is_file:
-            raise errors.FileExpected(path)
+        if self.strict:
+            info = self.getinfo(path)
+            if not info.is_file:
+                raise errors.FileExpected(path)
         _path = self.validatepath(path)
         _key = self._path_to_key(_path)
         bytes_file = io.BytesIO()
@@ -663,18 +686,19 @@ class S3FS(FS):
 
     def setbytes(self, path, contents):
         if not isinstance(contents, bytes):
-            raise ValueError('contents must be bytes')
+            raise TypeError('contents must be bytes')
 
         _path = self.validatepath(path)
         _key = self._path_to_key(_path)
-        if not self.isdir(dirname(path)):
-            raise errors.ResourceNotFound(path)
-        try:
-            info = self.getinfo(path)
-            if info.is_dir:
-                raise errors.FileExpected(path)
-        except errors.ResourceNotFound:
-            pass
+        if self.strict:
+            if not self.isdir(dirname(path)):
+                raise errors.ResourceNotFound(path)
+            try:
+                info = self.getinfo(path)
+                if info.is_dir:
+                    raise errors.FileExpected(path)
+            except errors.ResourceNotFound:
+                pass
 
         bytes_file = io.BytesIO(contents)
         with s3errors(path):
@@ -685,14 +709,16 @@ class S3FS(FS):
     def setbinfile(self, path, file):
         _path = self.validatepath(path)
         _key = self._path_to_key(_path)
-        if not self.isdir(dirname(path)):
-            raise errors.ResourceNotFound(path)
-        try:
-            info = self.getinfo(path)
-            if info.is_dir:
-                raise errors.FileExpected(path)
-        except errors.ResourceNotFound:
-            pass
+
+        if self.strict:
+            if not self.isdir(dirname(path)):
+                raise errors.ResourceNotFound(path)
+            try:
+                info = self.getinfo(path)
+                if info.is_dir:
+                    raise errors.FileExpected(path)
+            except errors.ResourceNotFound:
+                pass
 
         with s3errors(path):
             self.client.upload_fileobj(file, self._bucket_name, _key)
@@ -702,19 +728,25 @@ class S3FS(FS):
             raise errors.DestinationExists(dst_path)
         _src_path = self.validatepath(src_path)
         _dst_path = self.validatepath(dst_path)
-        if not self.isdir(dirname(_dst_path)):
-            raise errors.ResourceNotFound(dst_path)
+        if self.strict:
+            if not self.isdir(dirname(_dst_path)):
+                raise errors.ResourceNotFound(dst_path)
         _src_key = self._path_to_key(_src_path)
         _dst_key = self._path_to_key(_dst_path)
-        with s3errors(src_path):
-            self.client.copy_object(
-                Bucket=self._bucket_name,
-                Key=_dst_key,
-                CopySource={
-                    'Bucket':self._bucket_name,
-                    'Key':_src_key
-                }
-            )
+        try:
+            with s3errors(src_path):
+                self.client.copy_object(
+                    Bucket=self._bucket_name,
+                    Key=_dst_key,
+                    CopySource={
+                        'Bucket':self._bucket_name,
+                        'Key':_src_key
+                    }
+                )
+        except errors.ResourceNotFound:
+            if self.exists(src_path):
+                raise errors.FileExpected(src_path)
+            raise
 
     def move(self, src_path, dst_path, overwrite=False):
         self.copy(src_path, dst_path, overwrite=overwrite)
